/**
 * A Converter Function
 */
interface ConverterFunction<Result, Input = unknown> {
    readonly displayName?: string;
    /**
     * @param input - The Input value to convert
     * @param path - The path to the input value from the root entity
     * @param entity - The root entity that is being converted
     */
    (input: Input, path: (string | number)[], entity: unknown): Result;
}
/**
 * A standard Converter function with extended methods
 */
interface Converter<Result, Input = unknown> extends ConverterFunction<Result, Input> {
    /**
     * The Name of the converter
     */
    readonly displayName: string;
    /**
     * @param input - The Input value to convert
     * @param path - The path to the input value from the root entity, defaults to []
     * @param entity - The root entity, defaults to input
     */
    (input: Input, path?: (string | number)[], entity?: unknown): Result;
    /**
     * pipes the output of this converter to another converter
     * @param nextConverter - the next converter to call
     * @param name - the name of the resulting converter
     */
    pipe<Next>(nextConverter: ConverterFunction<Next, Result>, name?: string): Converter<Next, Input>;
    /**
     * if this conversion would throw an error
     * try the other conversion
     * @param otherConverter - the other conversion to try
     */
    or<Other>(otherConverter: ConverterFunction<Other, Input>): Converter<Result | Other, Input>;
    /**
     * Provide a default value fallback if the value is undefined
     * The given converter is invoked if the input value is undefined
     * otherwise this converter is invoked. The result of the
     * given converter is passed through this converter as well.
     * @param defaultConverter - the converter to select a default value
     * @returns a new converter that when given an undefined input tries
     * to supply a default value to this converter.
     */
    default(defaultConverter: ConverterFunction<Input, undefined>): Converter<Result, Input | undefined>;
}
/**
 * Given a converter function get the name of the converter
 * @param converter - the converter function
 */
declare function getConverterName(converter: ConverterFunction<any, any>): string;
/**
 * Given a converter function or converter extract the result type.
 */
declare type TypeOf<C extends Converter<any, any> | ConverterFunction<any, any>> = C extends Converter<infer R, any> ? R : C extends ConverterFunction<infer FR, any> ? FR : never;
/**
 * Given a converter function and optional name create a converter
 * @param converter - the converter function to use to convert the values
 * @param name - the name, if not specified uses the display name or name property of the converter
 */
declare function createConverter<Result, Input = unknown>(converter: ConverterFunction<Result, Input>, name?: string): Converter<Result, Input>;

/**
 * Converts values to numbers
 * This expects incoming values to be numbers or string then
 * calls the Number constructor to try to convert values
 */
declare const NumberConverter: Converter<number, unknown>;
/**
 * Converts defined non-object and non-functions to strings.
 * This calls the String constructor to convert incoming values
 */
declare const StringConverter: Converter<string, unknown>;
/**
 * Converts values to booleans
 * this accepts the string literals of "true" or "false" but otherwise expects a boolean or Boolean
 */
declare const BooleanConverter: Converter<boolean, unknown>;
/**
 * Passes through an unknown value unchanged.
 * Useful when used with the record container
 */
declare const UnknownConverter: Converter<unknown, unknown>;
/**
 * Always throws an error no mater what value is passed
 * Useful when dealing with deprecated properties that should not be provided
 */
declare const NeverConverter: Converter<never, unknown>;
/**
 * Given a value create a converter that must match the literal value
 * @param value - the literal value to compare to, uses === identity equality
 */
declare function literal<V extends string | number | boolean | null | undefined>(value: V): Converter<V, unknown>;
/**
 * Converter that matches the literal null,
 * You can make values optionally nullable by using
 * t.null.or(...)
 */
declare const NullConverter: Converter<null, unknown>;
/**
 * Converter that matches the literal undefined,
 * You can make values optionally undefined by using
 * t.undefined.or(...)
 */
declare const UndefinedConverter: Converter<undefined, unknown>;
/**
 * A type representing undefined or null
 */
declare type None = null | undefined;
/**
 * Converter that matches the literal undefined or null,
 * You can make values optionally none by using
 * t.none.or(...)
 */
declare const NoneConverter: Converter<null | undefined, unknown>;

/**
 * A Record\<string,T\> converter
 * @param converter - the converter to use for the values in the record
 */
declare function record<T>(converter: ConverterFunction<T, unknown>): Converter<Record<string, T>, unknown>;
/**
 * A Array\<T\> converter
 * @param converter - the converter to use for values of the array.
 */
declare function array<T>(converter: ConverterFunction<T, unknown>): Converter<T[], unknown>;

/**
 * Given a set of converters if one succeeds at converting the input return the
 * given value.
 *
 * @param converters - Array of converters to pass in.
 * @param value - Value to return
 *
 */
declare const coerce: <V>(converters: [Converter<any>, ...Converter<any>[]], value: V) => Converter<V, unknown>;
/**
 * A coercer that coerces values of None type to a null literal
 */
declare const noneAsNull: Converter<None>;
/**
 * A coercer that coerces values of None type to an undefined literal
 */
declare const noneAsUndefined: Converter<None>;

declare type ComposedConverters<T> = {
    [K in keyof T]: ConverterFunction<T[K]>;
};
/**
 * Combine an arbitrary number of converters and pass their resulting
 *   values into the `combiner` function.
 *
 * @param converters - Array of converters to pass in.
 * @param combiner - Function to call with the results of all the
 *   converter functions.
 */
declare const compose: <C extends (...args: any) => any>(converters: ComposedConverters<Parameters<C>>, combiner: C) => Converter<ReturnType<C>, unknown>;

/**
 * Given a converter function make an optional converter function
 * Optional functions allow undefined values to pass-through
 * @param converter - the inner converter function
 */
declare function optional<Result, Input = unknown>(converter: ConverterFunction<Result, Input>): Converter<Result | undefined, Input | undefined>;
/**
 * Given a converter function make an noneable converter that allows `none`
 * types to pass through
 * @param converter - the inner converter function
 */
declare function noneable<Result, Input = unknown>(converter: ConverterFunction<Result, Input>): Converter<Result | None, Input | None>;
/**
 * Given a converter function make an noneable converter function that also coerces
 * `none` to `null`
 * @param converter - the inner converter function
 */
declare function noneableAsNull<Result, Input = unknown>(converter: ConverterFunction<Result, Input>): Converter<Result | null, Input | None>;
/**
 * Given a converter function make an noneable converter function that also coerces
 * `none` to `undefined`
 * @param converter - the inner converter function
 */
declare function noneableAsUndefined<Result, Input = unknown>(converter: ConverterFunction<Result, Input>): Converter<Result | undefined, Input | None>;

/**
 * Converter Errors are the standard errors
 * coming from type-converters
 */
declare class ConverterError extends Error {
    /**
     * Error Fields records the path to fields that have an error.
     * The key of the errorFields is a path in the form of $.part[index]
     */
    readonly errorFields: Record<string, {
        expected: string;
        actual: unknown;
    }>;
    constructor(actual: unknown, expected: string, path: (string | number)[]);
}

/**
 * Given a path array return a path string
 * @param path - the path array comprising of string and number parts
 */
declare function formatPath(path: (string | number)[]): string;
/**
 * Determine the printed string for a value
 * @param value - a value to print
 */
declare function displayValue(value: unknown): string;

/**
 * When composing functions it is useful to expose a function that doesn't take additional arguments
 * @param converter - a converter to run on the input
 * @returns a function that passes its first argument only to the converter.
 */
declare const convert: <Result, Input>(converter: ConverterFunction<Result, Input>) => (input: Input) => Result;

/**
 * The ShapeConverter is a converter, however we expose the
 * converters that were used to create it. This allows for these
 * converters to be spread, in order to inherit them.
 */
interface ShapeConverter<S extends object> extends Converter<S, unknown> {
    readonly strict: boolean;
    readonly converters: {
        [K in keyof S]: Converter<S[K], unknown>;
    };
}
/**
 * A strict converter where properties not explicitly specified will not be returned
 * @param converters - an object structure describing the resulting value
 */
declare function strict<S extends object>(converters: {
    [K in keyof S]: ConverterFunction<S[K], unknown>;
}): ShapeConverter<S>;
/**
 * A loose shape converter that allows unknown properties to pass through
 * Generally shape should be used if the value is going to be "re-serialized" as-is
 * This is useful for things like application stacks which may want to verify a subset of
 * properties but leave most validation up to the domain stack
 * @param converters - an object structure describing the resulting value
 */
declare function shape<S extends object>(converters: {
    [K in keyof S]: ConverterFunction<S[K], unknown>;
}): ShapeConverter<S>;
/**
 * Given a shape converter turns all converters into optionals.
 *
 * If the given shape converter was created with the t.strict creator
 * the resulting partial will also be strict
 * @param converter - the shape converter to make partial
 */
declare function partial<S extends object>(converter: ShapeConverter<S>): ShapeConverter<{
    [K in keyof S]: S[K] | undefined;
}>;

declare const CurrentPath: unique symbol;
declare const ParentPath: unique symbol;
declare type TargetPathPart = string | number | typeof CurrentPath | typeof ParentPath;
/**
 * Given a Path and a converter, invoke the converter at the given path
 * from the root entity.
 *
 * This is especially useful for migrating data from old schemas
 * @param targetPath -the path to pull data from
 */
declare function forPath(targetPath: TargetPathPart[]): Converter<unknown, unknown>;
/**
 * Given a Path and a converter, invoke the converter at the given path
 * from the root entity.
 *
 * This is especially useful for migrating data from old schemas
 * @param targetPath - the path to pull data from
 * @param converter - the converter to process the data with
 */
declare function forPath<Result>(targetPath: TargetPathPart[], converter: ConverterFunction<Result, unknown>): Converter<Result, unknown>;
/**
 * Reroot the given converter at its current point in the tree.
 * This effectively causes functions like forPath to
 * be unable to traverse above the current point in the tree. It's useful as a way
 * to re-use converters when you have nested data.
 * @param converter - the converter to process data with
 */
declare function sub<Result, Input = unknown>(converter: ConverterFunction<Result, Input>): Converter<Result, Input>;

/**
 * One of a set of constant types
 * This uses triple equality (===) to compare
 */
declare function oneOf<V extends string | number | boolean | undefined | null>(options: V[]): Converter<V, unknown>;
/**
 * A union type for tagged unions
 * @param tag - the branch selector selects which union branch to use
 * @param options - the union branches
 */
declare function taggedUnion<R extends object, Input = unknown>(tag: keyof R, options: Record<string, ConverterFunction<R, Input>>): Converter<R, Input>;
/**
 * Flattens a converter function that returns a converter function into
 *   just a converter function.
 *
 * @param converter - Converter function to flatten.
 */
declare const select: <TOutput, TInput = unknown>(converter: ConverterFunction<ConverterFunction<TOutput, TInput>, TInput>) => Converter<TOutput, TInput>;

export { ComposedConverters, Converter, ConverterError, ConverterFunction, CurrentPath, None, ParentPath, ShapeConverter, TargetPathPart, TypeOf, array, BooleanConverter as boolean, coerce, compose, convert, createConverter, displayValue, forPath, formatPath, getConverterName, literal, NeverConverter as never, NoneConverter as none, noneAsNull, noneAsUndefined, noneable, noneableAsNull, noneableAsUndefined, NullConverter as null, NumberConverter as number, oneOf, optional, partial, record, select, shape, strict, StringConverter as string, sub, taggedUnion, UndefinedConverter as undefined, UnknownConverter as unknown };
